/**
 *  stardate.c: program to calculate the current stardate
 *  by Andrew Main <zefram@dcs.warwick.ac.uk>
 *  1995-12-21, stardate [-31]6778.27
 *
 *  This has been written to compile under traditional C.
 *
 *  This program will display the current stardate to two decimal places, or
 *  whatever is specified on the command line.  The output will change every
 *  three minutes (actually 172.8s) with the default two decimal places.
**/

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>

int stardate();

static int usage(msg)
char *msg;
{
  fputs(msg, stderr);
  fputs("Usage: stardate [ places ]\n", stderr);
  fputs("  `places' must be an integer in the range 0 to 6.", stderr);
  fputs("  The default is 2.\n", stderr);
  exit(1);
}

int main(argc, argv)
int argc;
char *argv[];
{
    char *ptr;
    long places;
    time_t tm;
    long issue, integer, fraction;
    char fractionbuffer[7];
    /* Parse command-line arguments.  The first argument specifies the *
     * number of decimal places to output (0-6).  Default is 2.        */
    places = 2;
    if(argc > 2)
	usage("Too many arguments\n");
    if(argc == 2) {
	places = strtol(argv[1], &ptr, 10);
	if(!*argv[1] || *ptr)
	    usage("Non-numeric argument\n");
	if(places<0 || places>6)
	    usage("Argument out of range\n");
    }
    /* Get current time in UNIX format. */
    tm = time((time_t *)NULL);
    if(tm == (time_t)-1) {
	fprintf(stderr, "Error in time()\n");
	exit(1);
    }
    /* Get stardate from stardate() function. */
    stardate(tm, &issue, &integer, &fraction);
    /* Output stardate, with requested number of decimal places. */
    if(places == 0)
	printf("[%ld]%04ld\n", issue, integer);
    else {
	sprintf(fractionbuffer, "%06ld", fraction);
	fractionbuffer[places] = 0;
	printf("[%ld]%04ld.%s\n", issue, integer, fractionbuffer);
    }
    /* Success. */
    exit(0);
}

/**
 *  stardate(): calculates stardate
 *  Parameters: time_t tm;          UNIX-type time
 *              long *issue;        returns issue number
 *              long *integer;      returns integral part of stardate
 *              long *fraction;     returns fractional part of stardate
 *
 *  It is also assumed that the date is before 2270-01-26, but this is after
 *  the end of time on UNIX systems, so it is not a problem.
 *
 *  The fraction returned is scaled so that it is in the range 000000 to
 *  999999 inclusive.  Its units digit represents 0.000001 stardate units.
 *  This digit theoretically changes every 0.01728 (exactly) seconds, but of
 *  course not every value can actually be returned.  This is because the time
 *  passed is only specified to a precision of one second.  The intent in
 *  making the function provide six fractional digits is to make the returned
 *  stardate different every second.
 *
 *  The returned values are not rounded at all, but are instead deliberately
 *  truncated.  This behaviour is desired, so that the date that is returned
 *  is known to be before or equal to the date passed; this is how times and
 *  dates are normally treated.
 *
 *  The function uses the knowledge that 00:00 UTC on 1970-01-01, the origin
 *  for UNIX time, is SD [-36]9350.  It also `knows' that each stardate unit
 *  covers 17280 seconds.
**/

int stardate(tm, issue, integer, fraction)
time_t tm;
long *issue, *integer, *fraction;
{
    /* It would be convenient to calculate the fractional part with *
     * *fraction = ( (tm%17280L) * 1000000 ) / 17280;               *
     * but the long int type may not be long enough for this (it    *
     * requires 36 bits).  Cancelling the 1000000 with the 17280    *
     * gives an expression that takes only 27 bits.                 */
    *fraction = ( (tm%17280L) * 3125 ) / 54;
    /* Get integer part. */
    *integer = tm/17280 + 9350;
    /* At this stage, *integer contains the issue number in the obvious *
     * place, biased to always be non-negative.  The issue number can   *
     * be extracted by simply dividing *integer by 10000 and offsetting *
     * it appropriately:                                                */
    *issue = ( (*integer) / 10000 ) - 36;
    /* Remove the issue number from *integer. */
    *integer %= 10000;
    /* Successful return. */
    return 0;
}
